@@ -6,7 +6,9 @@ Functions devoted to requests over the WWW.
 ## under GNU GPL v3 license
 
 import logging
-import urllib3
+
+import urllib3  # type: ignore
+
 
 RETRY_STRATEGY = urllib3.util.Retry(
     total=5,
@@ -19,8 +21,8 @@ ACCEPTABLE_CODES = {200, 300, 301, 302, 303, 304, 305, 306, 307, 308}
 
 
 # Test redirects
-def redirection_test(url):
-    """ Test final URL to handle redirects
+def redirection_test(url: str) -> str:
+    """Test final URL to handle redirects
     Args:
         url: url to check
 
@@ -30,18 +32,17 @@ def redirection_test(url):
     Raises:
         Nothing.
     """
-    #headers.update({
+    # headers.update({
     #    "User-Agent" : str(sample(settings.USER_AGENTS, 1)), # select a random user agent
-    #})
+    # })
     try:
-        rhead = HTTP_POOL.request('HEAD', url)
+        rhead = HTTP_POOL.request("HEAD", url)
     except Exception as err:
-        logging.error('unknown: %s %s', url, err) # sys.exc_info()[0]
+        logging.error("unknown: %s %s", url, err)  # sys.exc_info()[0]
     else:
         # response
         if rhead.status in ACCEPTABLE_CODES:
-            logging.debug('result found: %s %s', rhead.geturl(), rhead.status)
-            return rhead.geturl()
-    #else:
-    logging.debug('no result found: %s', url)
-    return None
+            logging.debug("result found: %s %s", rhead.geturl(), rhead.status)
+            return rhead.geturl()  # type: ignore
+    # else:
+    raise ValueError("cannot reach URL: %s", url)
