@@ -11,6 +11,9 @@ import sys
 from collections import defaultdict, deque
 from datetime import datetime, timedelta
 from threading import Lock
+from typing import Any, DefaultDict, Deque, Dict, List, Optional, Tuple, Union
+
+from urllib.robotparser import RobotFileParser
 
 from .filters import lang_filter, validate_url
 from .urlutils import get_host_and_path, is_known_link
@@ -21,80 +24,108 @@ LOGGER = logging.getLogger(__name__)
 
 class DomainEntry:
     "Class to record host-related information and URL paths."
-    __slots__ = ('all_visited', 'count', 'rules', 'timestamp', 'tuples')
+    __slots__ = ("all_visited", "count", "rules", "timestamp", "tuples")
 
-    def __init__(self):
-        self.all_visited = False
-        self.count = 0
-        self.rules = None
-        self.tuples = deque()
-        self.timestamp = None
+    def __init__(self) -> None:
+        self.all_visited: bool = False
+        self.count: int = 0
+        self.rules: Optional[RobotFileParser] = None
+        self.tuples: Deque[UrlPathTuple] = deque()
+        self.timestamp: Optional[Any] = None
 
 
 class UrlPathTuple:
     "Class storing information for URL paths relative to a domain/host."
-    __slots__ = ('urlpath', 'visited')
+    __slots__ = ("urlpath", "visited")
 
-    def __init__(self, urlpath, visited):
-        self.urlpath = urlpath
-        self.visited = visited
+    def __init__(self, urlpath: str, visited: bool) -> None:
+        self.urlpath: str = urlpath
+        self.visited: bool = visited
 
 
 class UrlStore:
     "Defines a class to store domain-classified URLs and perform checks against it."
-    __slots__ = ('compressed', 'done', 'language', 'strict', 'urldict', '_lock')
+    __slots__ = ("compressed", "done", "language", "strict", "urldict", "_lock")
 
-    def __init__(self, compressed=False, language=None, strict=False):
-        self.compressed = compressed
-        self.done = False
-        self.language = language
-        self.strict = strict
-        self.urldict = defaultdict(DomainEntry)
-        self._lock = Lock()
+    def __init__(
+        self,
+        compressed: bool = False,
+        language: Optional[str] = None,
+        strict: bool = False,
+    ) -> None:
+        self.compressed: bool = compressed
+        self.done: bool = False
+        self.language: Optional[str] = language
+        self.strict: bool = strict
+        self.urldict: DefaultDict[str, DomainEntry] = defaultdict(DomainEntry)
+        self._lock: Lock = Lock()
 
-        def dump_unvisited_urls(num, frame):
-            LOGGER.warning('Processing interrupted, dumping unvisited URLs from %s hosts', len(self.urldict))
+        def dump_unvisited_urls(num: Any, frame: Any) -> None:
+            LOGGER.warning(
+                "Processing interrupted, dumping unvisited URLs from %s hosts",
+                len(self.urldict),
+            )
             for domain in self.urldict:
-                print('\n'.join([domain + u.urlpath for u in self._load_urls(domain) if u.visited is False]), file=sys.stderr)
+                print(
+                    "\n".join(
+                        [
+                            domain + u.urlpath
+                            for u in self._load_urls(domain)
+                            if u.visited is False
+                        ]
+                    ),
+                    file=sys.stderr,
+                )
             sys.exit()
 
         # don't use the following on Windows
-        if not sys.platform.startswith('win'):
+        if not sys.platform.startswith("win"):
             signal.signal(signal.SIGINT, dump_unvisited_urls)
             signal.signal(signal.SIGTERM, dump_unvisited_urls)
 
-    #def _filter_urlpaths(self, domain, urls):
+    # def _filter_urlpaths(self, domain, urls):
     #    if self.validation is True or self.language is not None:
     #        return [u for u in urls if self._filter_url(domain + u) is True]
     #    return urls
 
-    def _buffer_urls(self, data, visited=False):
-        inputdict = defaultdict(deque)
+    def _buffer_urls(
+        self, data: List[str], visited: bool = False
+    ) -> DefaultDict[str, Deque[UrlPathTuple]]:
+        inputdict: DefaultDict[str, Deque[UrlPathTuple]] = defaultdict(deque)
         for url in list(dict.fromkeys(data)):
             # segment URL and add to domain dictionary
             try:
                 # validate
                 validation_result, parsed_url = validate_url(url)
                 if validation_result is False:
-                    LOGGER.debug('Invalid URL: %s', url)
+                    LOGGER.debug("Invalid URL: %s", url)
                     raise ValueError
                 # filter
-                if self.language is not None and lang_filter(url, self.language, self.strict) is False:
-                    LOGGER.debug('Wrong language: %s', url)
+                if (
+                    self.language is not None
+                    and lang_filter(url, self.language, self.strict) is False
+                ):
+                    LOGGER.debug("Wrong language: %s", url)
                     raise ValueError
                 hostinfo, urlpath = get_host_and_path(parsed_url)
                 inputdict[hostinfo].append(UrlPathTuple(urlpath, visited))
             except (TypeError, ValueError):
-                LOGGER.warning('Discarding URL: %s', url)
+                LOGGER.warning("Discarding URL: %s", url)
         return inputdict
 
-    def _load_urls(self, domain):
+    def _load_urls(self, domain: str) -> Deque[UrlPathTuple]:
         value = self.urldict[domain].tuples
         if isinstance(value, bytes):
             return pickle.loads(bz2.decompress(value))
         return value
 
-    def _store_urls(self, domain, to_right=None, timestamp=None, to_left=None):
+    def _store_urls(
+        self,
+        domain: str,
+        to_right: Optional[Deque[UrlPathTuple]] = None,
+        timestamp: Optional[datetime] = None,
+        to_left: Optional[Deque[UrlPathTuple]] = None,
+    ) -> None:
         if domain in self.urldict:
             urls = self._load_urls(domain)
             known = {u.urlpath for u in urls}
@@ -110,7 +141,9 @@ class UrlStore:
         with self._lock:
             # compression
             if self.compressed is True:
-                self.urldict[domain].tuples = bz2.compress(pickle.dumps(urls, protocol=4))
+                self.urldict[domain].tuples = bz2.compress(  # type: ignore
+                    pickle.dumps(urls, protocol=4)
+                )
             else:
                 self.urldict[domain].tuples = urls
             # adjust all_visited status
@@ -119,9 +152,12 @@ class UrlStore:
             if timestamp is not None:
                 self.urldict[domain].timestamp = timestamp
 
-    def _search_urls(self, urls, switch=None):
+    def _search_urls(
+        self, urls: List[str], switch: Optional[int] = None
+    ) -> List[Union[Any, str]]:
         # init
-        last_domain, known_paths = None, set()
+        last_domain: Optional[str] = None
+        known_paths: Dict[str, Optional[bool]] = {}
         remaining_urls = {u: None for u in urls}
         # iterate
         for url in sorted(remaining_urls):
@@ -130,9 +166,11 @@ class UrlStore:
             if hostinfo != last_domain:
                 last_domain = hostinfo
                 if switch == 1:
-                    known_paths = {u.urlpath for u in self._load_urls(hostinfo)}
+                    known_paths = {u.urlpath: None for u in self._load_urls(hostinfo)}
                 elif switch == 2:
-                    known_paths = {u.urlpath: u.visited for u in self._load_urls(hostinfo)}
+                    known_paths = {
+                        u.urlpath: u.visited for u in self._load_urls(hostinfo)
+                    }
             # run checks: case 1: the path matches, case 2: visited URL
             if urlpath in known_paths and (
                 switch == 1 or (switch == 2 and known_paths[urlpath] is True)
@@ -141,12 +179,17 @@ class UrlStore:
         # preserve input order
         return list(remaining_urls)
 
-    def _timestamp(self, domain):
+    def _timestamp(self, domain: str) -> Optional[datetime]:
         return self.urldict[domain].timestamp
 
     # URL MANIPULATION AND INFO
 
-    def add_urls(self, urls=None, appendleft=None, visited=False):
+    def add_urls(
+        self,
+        urls: Optional[List[str]] = None,
+        appendleft: Optional[List[str]] = None,
+        visited: bool = False,
+    ) -> None:
         """Add a list of URLs to the (possibly) existing one.
         Optional: append certain URLs to the left,
         specify if the URLs have already been visited."""
@@ -157,35 +200,35 @@ class UrlStore:
             for host, urltuples in self._buffer_urls(appendleft, visited).items():
                 self._store_urls(host, to_left=urltuples)
 
-    def is_known(self, url):
+    def is_known(self, url: str) -> bool:
         "Check if the given URL has already been stored."
         hostinfo, urlpath = get_host_and_path(url)
         values = self._load_urls(hostinfo)
         # returns False if domain or URL is new
         return urlpath in {u.urlpath for u in values}
 
-    def find_known_urls(self, domain):
+    def find_known_urls(self, domain: str) -> List[str]:
         """Get all already known URLs for the given domain (ex. "https://example.org")."""
         values = self._load_urls(domain)
         return [domain + u.urlpath for u in values]
 
-    def filter_unknown_urls(self, urls):
+    def filter_unknown_urls(self, urls: List[str]) -> List[str]:
         "Take a list of URLs and return the currently unknown ones."
         return self._search_urls(urls, switch=1)
 
-    def get_known_domains(self):
+    def get_known_domains(self) -> List[str]:
         "Return all known domains as a list."
         return list(self.urldict)
 
-    def is_exhausted_domain(self, domain):
+    def is_exhausted_domain(self, domain: str) -> bool:
         "Tell if all known URLs for the website have been visited."
         if domain in self.urldict:
             return self.urldict[domain].all_visited
-        raise KeyError('website not in store')
+        raise KeyError("website not in store")
 
     # URL-BASED QUERIES
 
-    def has_been_visited(self, url):
+    def has_been_visited(self, url: str) -> bool:
         "Check if the given URL has already been visited.."
         hostinfo, urlpath = get_host_and_path(url)
         values = self._load_urls(hostinfo)
@@ -193,22 +236,22 @@ class UrlStore:
         # defaults to None, thus False
         return known_urlpaths.get(urlpath) or False
 
-    def find_unvisited_urls(self, domain):
+    def find_unvisited_urls(self, domain: str) -> List[str]:
         "Get all unvisited URLs for the given domain."
         values = self._load_urls(domain)
         return [domain + u.urlpath for u in values if u.visited is False]
 
-    def filter_unvisited_urls(self, urls):
+    def filter_unvisited_urls(self, urls: List[str]) -> List[Union[Any, str]]:
         "Take a list of URLs and return the currently unvisited ones."
         return self._search_urls(urls, switch=2)
 
-    def unvisited_websites_number(self):
+    def unvisited_websites_number(self) -> int:
         "Return the number of websites for which there are still URLs to visit."
         return len([d for d in self.urldict if self.urldict[d].all_visited is False])
 
     # DOWNLOADS
 
-    def get_url(self, domain):
+    def get_url(self, domain: str) -> Optional[str]:
         "Retrieve a single URL and consider it to be visited (with corresponding timestamp)."
         # not fully used
         if self.urldict[domain].all_visited is False:
@@ -226,42 +269,54 @@ class UrlStore:
             self.urldict[domain].all_visited = True
         return None
 
-    def get_download_urls(self, timelimit=10):
+    def get_download_urls(self, timelimit: int = 10) -> Optional[List[str]]:
         """Get a list of immediately downloadable URLs according to the given
-           time limit per domain."""
+        time limit per domain."""
         with self._lock:
-            potential = [d for d in self.urldict if self.urldict[d].all_visited is False]
+            potential = [
+                d for d in self.urldict if self.urldict[d].all_visited is False
+            ]
         if not potential:
             self.done = True
             return None
         targets = []
         for domain in potential:
             timestamp = self._timestamp(domain)
-            if timestamp is None or (datetime.now() - timestamp).total_seconds() > timelimit:
+            if (
+                timestamp is None
+                or (datetime.now() - timestamp).total_seconds() > timelimit
+            ):
                 targets.append(domain)
         # get corresponding URLs and filter out None values
         return list(filter(None, [self.get_url(domain) for domain in targets]))
 
-    def establish_download_schedule(self, max_urls=100, time_limit=10):
+    def establish_download_schedule(
+        self, max_urls: int = 100, time_limit: int = 10
+    ) -> List[str]:
         """Get up to the specified number of URLs along with a suitable
-           backoff schedule (in seconds)."""
+        backoff schedule (in seconds)."""
         # see which domains are free
         with self._lock:
-            potential = [d for d in self.urldict if self.urldict[d].all_visited is False]
+            potential = [
+                d for d in self.urldict if self.urldict[d].all_visited is False
+            ]
             if not potential:
                 self.done = True
                 return []
         # variables init
         per_domain = max_urls // len(potential) or 1
-        targets = []
+        targets: List[Tuple[float, str]] = []
         # iterate potential domains
         for domain in potential:
             url_tuples = self._load_urls(domain)
             # load urls
-            urlpaths = []
+            urlpaths: List[str] = []
             # get first non-seen urls
             for url in url_tuples:
-                if len(urlpaths) >= per_domain or (len(targets) + len(urlpaths)) >= max_urls:
+                if (
+                    len(urlpaths) >= per_domain
+                    or (len(targets) + len(urlpaths)) >= max_urls
+                ):
                     break
                 if url.visited is False:
                     urlpaths.append(url.urlpath)
@@ -271,10 +326,15 @@ class UrlStore:
             # determine timestamps
             now = datetime.now()
             original_timestamp = self._timestamp(domain)
-            if original_timestamp is None or (now - original_timestamp).total_seconds() > time_limit:
-                schedule_secs = 0
+            if (
+                original_timestamp is None
+                or (now - original_timestamp).total_seconds() > time_limit
+            ):
+                schedule_secs = 0.0
             else:
-                schedule_secs = time_limit - float(f'{(now - original_timestamp).total_seconds():.2f}')
+                schedule_secs = time_limit - float(
+                    f"{(now - original_timestamp).total_seconds():.2f}"
+                )
             for urlpath in urlpaths:
                 targets.append((schedule_secs, domain + urlpath))
                 schedule_secs += time_limit
@@ -283,30 +343,37 @@ class UrlStore:
             # store new info
             self._store_urls(domain, url_tuples, timestamp=total_diff)
         # sort by first tuple element (time in secs)
-        return sorted(targets, key=lambda x: x[0])
+        return sorted(targets, key=lambda x: x[0])  # type: ignore[arg-type]
 
-    def get_rules(self, website):
+    def get_rules(self, website: str) -> Optional[RobotFileParser]:
         "Return the stored crawling rules for the given website."
         return self.urldict[website].rules
 
     # GENERAL INFO
 
-    def total_url_number(self):
+    def total_url_number(self) -> int:
         "Find number of all URLs in store."
         return sum(len(self.urldict[d].tuples) for d in self.urldict)
 
-    def download_threshold_reached(self, threshold):
+    def download_threshold_reached(self, threshold: float) -> bool:
         "Find out if the download limit (in seconds) has been reached for one of the websites in store."
         return any(self.urldict[d].count >= threshold for d in self.urldict)
 
-    def dump_urls(self):
+    def dump_urls(self) -> List[str]:
         "Return a list of all known URLs."
         urls = []
         for domain in self.get_known_domains():
             urls.extend(self.find_known_urls(domain))
         return urls
 
-    def print_urls(self):
+    def print_urls(self) -> None:
         "Print all URLs in store (URL + TAB + visited or not)."
         for domain in self.get_known_domains():
-            print('\n'.join([domain + u.urlpath + '\t' + str(u.visited) for u in self._load_urls(domain)]))
+            print(
+                "\n".join(
+                    [
+                        domain + u.urlpath + "\t" + str(u.visited)
+                        for u in self._load_urls(domain)
+                    ]
+                )
+            )
